# -*- coding: utf-8 -*-

from itertools import product


def cal_v_s(routes: list,
            state_spaces: list,
            gamma: float,
            first_visit: bool = False,
            backtrack: bool = True):
    '''
    | 根据采样路径列表routes估计状态价值
    | 路径列表routes的每个元素为一条路径（也是一个列表）
    | 每条路径的元素为(s|状态, a|动作, r|奖励)三元组
    '''
    n_s = {s: 0 for s in state_spaces}
    v_s = {s: 0 for s in state_spaces}
    if not first_visit:
        if backtrack:
            # 一条路径上从后往前计算状态价值
            for route in routes:
                k_r = 0
                for k in range(len(route)-1, -1, -1):
                    (s, a, r) = route[k]
                    k_r = r + gamma * k_r # 步骤(时间)k的奖励
                    n_s[s] = n_s[s] + 1
                    v_s[s] = v_s[s] + (k_r-v_s[s]) / n_s[s]
        else:
            # 一条路径上从前往后计算状态价值
            for route in routes:
                n_ = len(route)
                # 从前往后计算须先取出每个步骤(时间)的奖励
                r_route = [x[2] for x in route]
                for k in range(n_):
                    (s, a, r) = route[k]
                    # 步骤(时间)k的奖励
                    k_r = sum([r_route[_]*gamma**(_-k) for _ in range(k, n_)])
                    n_s[s] = n_s[s] + 1
                    v_s[s] = v_s[s] + (k_r-v_s[s]) / n_s[s]
    else:
        for route in routes:
            k_r = 0
            route_r = []
            for k in range(len(route)-1, -1, -1):
                (s, a, r) = route[k]
                k_r = r + gamma * k_r
                route_r.append((s, k_r))
            visited = {s: 0 for s in state_spaces}
            for s, k_r in route_r[::-1]:
                # 仅考虑首次访问
                if visited[s]:
                    break
                visited[s] = 1
                n_s[s] = n_s[s] + 1
                v_s[s] = v_s[s] + (k_r-v_s[s]) / n_s[s]        
    return v_s


def cal_v_sa(routes: list,
             state_spaces: list,
             action_spaces: list,
             gamma: float,
             first_visit: bool = False,
             backtrack: bool = True):
    '''
    根据采样路径列表routes估计状态-动作价值
    '''
    sa_spaces = list(product(state_spaces, action_spaces))
    n_sa = {sa: 0 for sa in sa_spaces}
    v_sa = {sa: 0 for sa in sa_spaces}
    if not first_visit:
        if backtrack:
            # 从后往前计算
            for route in routes:
                k_r = 0
                for k in range(len(route)-1, -1, -1):
                    (s, a, r) = route[k]
                    sa = (s, a)
                    k_r = r + gamma * k_r
                    n_sa[sa] = n_sa[sa] + 1
                    v_sa[sa] = v_sa[sa] + (k_r-v_sa[sa]) / n_sa[sa]
        else:
            # 从前往后计算
            for route in routes:
                n_ = len(route)
                r_route = [x[2] for x in route]
                for k in range(n_):
                    (s, a, r) = route[k]
                    sa = (s, a)
                    k_r = sum([r_route[_]*gamma**(_-k) for _ in range(k, n_)])
                    n_sa[sa] = n_sa[sa] + 1
                    v_sa[sa] = v_sa[sa] + (k_r-v_sa[sa]) / n_sa[sa]
    else:
        for route in routes:
            k_r = 0
            route_r = []
            for k in range(len(route)-1, -1, -1):
                (s, a, r) = route[k]
                sa = (s, a)
                k_r = r + gamma * k_r
                route_r.append((sa, k_r))
            visited = {sa: 0 for sa in sa_spaces}
            for sa, k_r in route_r[::-1]:
                if visited[sa]:
                    break
                visited[sa] = 1
                n_sa[sa] = n_sa[sa] + 1
                v_sa[sa] = v_sa[sa] + (k_r-v_sa[sa]) / n_sa[sa]        
    return v_sa


def cal_v_s_a(routes: list,
              state_spaces: list,
              action_spaces: list,
              gamma: float,
              first_visit: bool = False,
              backtrack: bool = True):
    '''
    根据采样路径列表routes估计状态-动作价值
    '''
    n_s_a = {s: {a: 0 for a in action_spaces} for s in state_spaces}
    v_s_a = {s: {a: 0 for a in action_spaces} for s in state_spaces}
    if not first_visit:
        if backtrack:
            for route in routes:
                k_r = 0
                for k in range(len(route)-1, -1, -1):
                    (s, a, r) = route[k]
                    k_r = r + gamma * k_r
                    n_s_a[s][a] = n_s_a[s][a] + 1
                    v_s_a[s][a] = v_s_a[s][a] + (k_r-v_s_a[s][a]) / n_s_a[s][a]
        else:
            for route in routes:
                n_ = len(route)
                r_route = [x[2] for x in route]
                for k in range(n_):
                    (s, a, r) = route[k]
                    k_r = sum([r_route[_]*gamma**(_-k) for _ in range(k, n_)])
                    n_s_a[s][a] = n_s_a[s][a] + 1
                    v_s_a[s][a] = v_s_a[s][a] + (k_r-v_s_a[s][a]) / n_s_a[s][a]
    else:
        for route in routes:
            k_r = 0
            route_r = []
            for k in range(len(route)-1, -1, -1):
                (s, a, r) = route[k]
                s_a = (s, a)
                k_r = r + gamma * k_r
                route_r.append((s_a, k_r))
            visited = {(s, a): 0 for a in action_spaces for s in state_spaces}
            for s_a, k_r in route_r[::-1]:
                if visited[s_a]:
                    break
                visited[s_a] = 1
                s, a = s_a
                n_s_a[s][a] = n_s_a[s][a] + 1
                v_s_a[s][a] = v_s_a[s][a] + (k_r-v_s_a[s][a]) / n_s_a[s][a]        
    return v_s_a


def v_sa_to_v_s(v_sa: dict, p_sa: dict):
    v_ = {sa: p_sa[sa]*v for sa, v in v_sa.items()}
    v_s = {s: 0 for (s, a) in v_.keys()}
    for (s, a), v in v_.items():
        v_s[s] += v
    return v_s


def v_s_a_to_v_s(v_s_a, p_s_a):
    v_s = {}
    for s in v_s_a:
        v_s[s] = sum([p_s_a[s][a]*v for a, v in v_s_a[s].items()])
    return v_s

#%%
def ob2state(observation):
    return observation[0], observation[1], int(observation[2])


def evaluate_action_monte_carlo(env, policy, episode_num=10000):
    q = np.zeros_like(policy)
    c = np.zeros_like(policy)
    routes = []
    for _ in range(episode_num):
        # 玩一回合
        state_actions = []
        observation, _ = env.reset()
        while True:
            state = ob2state(observation)
            # print(policy[state])
            # print(state)
            action = np.random.choice(env.action_space.n, p=policy[state])
            observation, reward, terminated, truncated, _ = env.step(action)
            state_actions.append((state, action, reward))
            if terminated or truncated:
                break # 回合结束
        routes.append(state_actions)
        g = reward # 回报
        for state, action, _ in state_actions:
            c[state][action] += 1.
            q[state][action] += (g - q[state][action]) / c[state][action]
    return q, routes


def plot(data):
    fig, axes = plt.subplots(1, 2, figsize=(9, 4))
    titles = ['without ace', 'with ace']
    have_aces = [0, 1]
    extent = [12, 22, 1, 11]
    for title, have_ace, axis in zip(titles, have_aces, axes):
        dat = data[extent[0]:extent[1], extent[2]:extent[3], have_ace].T
        axis.imshow(dat, extent=extent, origin='lower')
        axis.set_xlabel('player sum')
        axis.set_ylabel('dealer showing')
        axis.set_title(title)
        
        
def monte_carlo_with_exploring_start(env, episode_num=10000):
    policy = np.zeros((22, 11, 2, 2))
    policy[:, :, :, 1] = 1.
    q = np.zeros_like(policy)
    c = np.zeros_like(policy)
    for _ in range(episode_num):
        # 随机选择起始状态和起始动作
        state = (np.random.randint(12, 22),
                 np.random.randint(1, 11),
                 np.random.randint(2))
        action = np.random.randint(2)
        # 玩一回合
        env.reset()
        if state[2]: # 有A
            env.player = [1, state[0] - 11]
        else: # 没有A
            if state[0] == 21:
                env.player = [10, 9, 2]
            else:
                env.player = [10, state[0] - 10]
        env.dealer[0] = state[1]
        state_actions = []
        while True:
            state_actions.append((state, action))
            observation, reward, terminated, truncated, _ = env.step(action)
            if terminated or truncated:
                break # 回合结束
            state = ob2state(observation)
            action = np.random.choice(env.action_space.n, p=policy[state])
        g = reward # 回报
        for state, action in state_actions:
            c[state][action] += 1.
            q[state][action] += (g - q[state][action]) / c[state][action]
            a = q[state].argmax()
            policy[state] = 0.
            policy[state][a] = 1.
    return policy, q


def monte_carlo_with_soft(env, episode_num=10000, epsilon=0.1):
    policy = np.ones((22, 11, 2, 2)) * 0.5 # 柔性策略
    q = np.zeros_like(policy)
    c = np.zeros_like(policy)
    for _ in range(episode_num):
        # 玩一回合
        state_actions = []
        observation, _ = env.reset()
        while True:
            state = ob2state(observation)
            action = np.random.choice(env.action_space.n, p=policy[state])
            state_actions.append((state, action))
            observation, reward, terminated, truncated, _ = env.step(action)
            if terminated or truncated:
                break # 回合结束
        g = reward # 回报
        for state, action in state_actions:
            c[state][action] += 1.
            q[state][action] += (g - q[state][action]) / c[state][action]
            # 更新策略为柔性策略
            a = q[state].argmax()
            policy[state] = epsilon / 2.
            policy[state][a] += (1. - epsilon)
    return policy, q


if __name__ == '__main__':
    from dramkit import TimeRecoder
    tr = TimeRecoder()
    import numpy as np
    # np.random.seed(0)
    import matplotlib.pyplot as plt
    try:
        import gymnasium as gym
        # import gym
    except:
        import gym
    
    env = gym.make("Blackjack-v1")
    # print('观察空间 = {}'.format(env.observation_space))
    # print('动作空间 = {}'.format(env.action_space))
    # print('动作数量 = {}'.format(env.action_space.n))

    def play_once_rand(env):
        total_reward = 0
        observation, _ = env.reset()
        print('观测 = {}'.format(observation))
        print('玩家 = {}, 庄家 = {}'.format(env.player, env.dealer))
        while True:
            action = np.random.choice(env.action_space.n)
            print('动作 = {}'.format(action))
            observation, reward, terminated, truncated, _ = env.step(action)
            print('观测 = {}, 奖励 = {}, 结束指示 = {}, 截断指示 = {}'.format(
                    observation, reward, terminated, truncated))
            print('玩家 = {}, 庄家 = {}'.format(env.player, env.dealer))
            total_reward += reward
            if terminated or truncated:
                return total_reward # 回合结束
    # 随机策略玩一回合
    print("随机策略, 奖励：{}".format(play_once_rand(env)))
    
    # 生成一个策略
    policy = np.zeros((22, 11, 2, 2))
    policy[20:, :, :, 0] = 1 # >= 20 时收手
    policy[:20, :, :, 1] = 1 # < 20 时继续
    
    q, routes = evaluate_action_monte_carlo(env, policy) # 动作价值
    v = (q * policy).sum(axis=-1) # 状态价值
    
    plot(v)
    
    state_spaces = list(product(range(22), range(11), range(2)))
    
    gamma = 1.0
    # first_visit = True
    first_visit = False
    # backtrack = True
    backtrack = False
    v_s = cal_v_s(routes, state_spaces, gamma,
                  first_visit=first_visit, backtrack=backtrack)
    v_ = v.copy()
    for s in v_s:
        v_[s] = v_s[s]
        
    action_spaces = [0, 1]
    v_sa = cal_v_sa(routes, state_spaces, action_spaces, gamma,
                    first_visit=first_visit, backtrack=backtrack)
    q_ = q.copy()
    for sa in v_sa:
        q_[tuple(list(sa[0])+[sa[1]])] = v_sa[sa]
        
    v_s_a = cal_v_s_a(routes, state_spaces, action_spaces, gamma,
                      first_visit=first_visit, backtrack=backtrack)
    q_1 = q.copy()
    for s in v_s_a:
        for a in action_spaces:
            q_1[s][a] = v_s_a[s][a]
            if v_s_a[s][a] != v_sa[(s, a)]:
                print(s, a)
        
    sa_spaces = product(state_spaces, action_spaces)
    p_sa = {sa: 0 for sa in sa_spaces}
    p_s_a = {s: {a: 0 for a in action_spaces} for s in state_spaces}
    for (s1, s2, s3), a in p_sa.keys():
        if s1 < 20:
            if a == 1:
                p_sa[((s1, s2, s3), a)] = 1
                p_s_a[(s1, s2, s3)][a] = 1
        else:
            if a == 0:
                p_sa[((s1, s2, s3), a)] = 1
                p_s_a[(s1, s2, s3)][a] = 1
            
    v_s1 = v_sa_to_v_s(v_sa, p_sa)
    v_s2 = v_s_a_to_v_s(v_s_a, p_s_a)
    
    
    # policy, q = monte_carlo_with_exploring_start(env)
    # v = q.max(axis=-1)
    # plot(policy.argmax(-1))
    # plot(v)
    
    
    # policy, q = monte_carlo_with_soft(env)
    # v = q.max(axis=-1)
    # plot(policy.argmax(-1))
    # plot(v)
    
    
    tr.used()












